/* $LAAS: color.c 2008/06/30 11:08:16 mallet $
 *
 * Copyright (c) 2007-2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Wed Feb 21 2007
 */

#include <stdio.h>
#include <math.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
#endif


/* --- viam_color_geosetup ------------------------------------------- */

/** \brief Geometry for color filters that do not change geometry.
 *
 * \return 0 on success.
 */

int
viam_color_geosetup(viam_handle_t vh, viam_image_t image,
		    viam_filterargs_t *args, viam_geometry_t *geom, void **data)
{
  /* check validity */
  if (geom->channels != 3) {
    warnx("color filters require three channels image");
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  /* geometry is unchanged */
  return 0;
}


/* --- viam_awb ------------------------------------------------------ */

/** \brief White-balance correction
 *
 * Apply U (blue) and V (red) gain.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image of the filter.
 * \param[in] source	BGR image on which to perform analysis.
 * \param[out] result	Computed image.
 * \param[out] args	Gains
 * \param[in] data	Private data
 *
 * \return 0 on sucess.
 */


int
viam_awb(viam_handle_t vh, viam_image_t image, const IplImage *source,
	 IplImage *result, viam_filterargs_t *args, void *data)
{
  int i, j;
  unsigned char *sl, *sp, *dl, *dp;

  /* apply B-G and R-G gains */

  sl = source->imageData;
  sp = result->imageData;
  for(i=0; i<source->height; i++) {
    for(j=0, sp=sl, dp=dl; j<source->width; j++, sp+=3, dp+=3) {

#define clamp(v)	(((v)<0)?0:(((v)>255)?255:(v)))
      dp[0] = clamp(sp[0] * 2.*args->blue);
      dp[1] = sp[1];
      dp[2] = clamp(sp[2] * 2.*args->red);
#undef clamp
    }
    sl += source->widthStep;
    sp += result->widthStep;
  }

  return 0;
}


/* --- viam_awb_auto ------------------------------------------------- */

/** \brief Auto white-balance algorithm
 *
 * Compute U (blue) and V (red) gain so that red, green and blue
 * pixel average are equal.
 *
 * \param[in] vh	Library handle
 * \param[in] image	Image used for initial acquisition
 * \param[in] source	Source image
 * \param[out] aumode	Argument mode
 * \param[in] args	Filter arguments
 * \param[in] data	Filter private data
 *
 * \return 0 on sucess.
 */

int
viam_awb_auto(viam_handle_t vh, viam_image_t image, const IplImage *source,
	      viam_filterautomode_t *aumode, viam_filterargs_t *args, void *data)
{
  int i, j, k;
  unsigned char *l, *p;
  double pblue, pred;
  int avb, avg, avr;

  /* compute B, G and R averages for 1 pixel out of 100 */
  avb = avg = avr = 0;
  l = source->imageData;
  for(i=0; i<source->height/10; i++) {
    for(j=0, p=l; j<source->width/10; j++, p+=3*9) {
      avb += *(p++); avg += *(p++); avr += *(p++);
    }

    l += 9*source->widthStep;
  }
  k = i * j;

  /* compute new gains */
  pblue = args->blue;  pred = args->red;

#define clamp(v)	(((v)<0.)?0.:(((v)>1.)?1.:(v)))
  args->blue = clamp(args->blue + (double)(avg - avb)/k/1024.);
  args->red =  clamp(args->red + (double)(avg - avr)/k/1024.);
#undef clamp

  /* stop one push when stable */
  if (*aumode == VIAM_FILTER_SOFTWARE_ONE_PUSH)
    if (fabs(args->blue-pblue) < 1e-3 && fabs(args->red-pred) < 1e-3)
      *aumode = VIAM_FILTER_MANUAL;

  return 0;
}
